#!/usr/bin/env python
# -*- coding: utf-8; -*-

# 2018-02-26 David Goncalves <david@lestat.st>
#            Python class to communicate with an MPU6050 chip on I2C bus.
#            It's a TDK Invensense 6-axis gyroscope and accelerometer.

import math
import smbus
import sys
import time


class mpu6050:

    # Internal register addresses
    REG_GYRO_CONFIG = 0x1b
    REG_ACCEL_CONFIG = 0x1c

    REG_ACCEL_XOUT_H = 0x3b
    REG_ACCEL_XOUT_L = 0x3c
    REG_ACCEL_YOUT_H = 0x3d
    REG_ACCEL_YOUT_L = 0x3e
    REG_ACCEL_ZOUT_H = 0x3f
    REG_ACCEL_ZOUT_L = 0x40

    REG_TEMP_OUT_H = 0x41
    REG_TEMP_OUT_L = 0x42

    REG_GYRO_XOUT_H = 0x43
    REG_GYRO_XOUT_L = 0x44
    REG_GYRO_YOUT_H = 0x45
    REG_GYRO_YOUT_L = 0x46
    REG_GYRO_ZOUT_H = 0x47
    REG_GYRO_ZOUT_L = 0x48

    REG_FIFO_COUNT_H = 0x72
    REG_FIFO_COUNT_L = 0x73

    REG_WHO_AM_I = 0x75

    accel_fs_sel = 0                                          # Accel AFS_SEL
    accel_fs_sel_map = [2, 4, 8, 16]
    accel_fs_sel_map_lsb = [16384.0, 8192.0, 4096.0, 2048.0]

    gyro_fs_sel = 0                                           # Gyro FS_SEL
    gyro_fs_sel_map = [250, 500, 1000, 3000]                  # Full Scale Range mapper
    gyro_fs_sel_map_lsb = [131.0, 65.5, 32.8, 16.4]           # LSB sensitivity mapper

    def __init__(self, address=0x68, bus=1, debug=False):
        self._debug = debug
        self.i2c_addr = address
        self.i2c_bus = smbus.SMBus(bus)

        self._debug_message("Accessing MPU6050 at address 0x%x on bus %s" % (address, bus))
        self._read_config()

    def _debug_message(self, message="", error=False):
        """
        Prints debug messages to console when debuging mode is activated
        """

        if (self._debug is True) or (error is True):
            print("%s %s" % (("[DEBUG]" if error is False else "[ERROR]"), message))
            sys.stdout.flush()
            if (error is True):
                raise Exception(message)

    def _write_data(self, address, data):
        """
        Private method to write data to the MPU6050
        """

        self._debug_message("> Writing %s bytes to register 0x%x [%s]" % (len(data), address, ' '.join(["0x%02X" % x for x in data])))
        self.i2c_bus.write_i2c_block_data(self.i2c_addr, address, data)

    def _read_data(self, address, size):
        """
        Private method to read data from the MPU6050
        """

        self._debug_message("> Reading %s bytes from register 0x%x" % (size, address))
        data = self.i2c_bus.read_i2c_block_data(self.i2c_addr, address, size)
        self._debug_message("> MPU6050 response : [%s]" % ' '.join(["0x%02X" % x for x in data]))

        return(data)

    def _two_comp(self, value, bitWidth):
        """
        Return the two complement's value of a value using 'bitWidth' resolution.

        This function comes from the following post in StackOverflow
        https://stackoverflow.com/questions/1604464/twos-complement-in-python
        """

        if value >= 2**bitWidth:
            # This catches when someone tries to give a value that is out of range
            raise ValueError("Value: {} out of range of {}-bit value.".format(value, bitWidth))
        else:
            return value - int((value << 1) & 2**bitWidth)

    def _read_config(self):
        """
        Reads gyroscope and accelerometer config to update local vars
        """

        self._debug_message("Reading Accel and Gyro config")
        data = self._read_data(self.REG_GYRO_CONFIG, 2)       # Read two registers at once

        # Set LSB sensitivity according to FS_SEL
        self.gyro_fs_sel = (data[0] >> 3) & 0x3
        self.gyro_lsb_sensitivity = self.gyro_fs_sel_map_lsb[self.gyro_fs_sel]
        self._debug_message("> Setting Gyro sensitivity to '%s°/s' (%s LSB/°/s)" % (self.gyro_fs_sel_map[self.gyro_fs_sel], self.gyro_lsb_sensitivity))

        # Set LSB sensitivity according to AFS_SEL
        self.accel_fs_sel = (data[1] >> 3) & 0x3
        self.accel_lsb_sensitivity = self.accel_fs_sel_map_lsb[self.accel_fs_sel]
        self._debug_message("> Setting Accel sensitivity to '%sg' (%s LSB/g)" % (self.accel_fs_sel_map[self.accel_fs_sel], self.accel_lsb_sensitivity))

    def GetFIFOStatus(self):
        """
        Reads the number of bytes in the FIFO buffer of the sensor
        """

        self._debug_message("Reading FIFO count registers")
        data = self._read_data(self.REG_FIFO_COUNT_H, 2)

    def GetTemperature(self):
        """
        Reads the temperature register of the sensor
        """

        self._debug_message("Reading sensor temperature")
        data = self._read_data(self.REG_TEMP_OUT_H, 2)
        temp = (self._two_comp((data[0] << 8) + data[1], 16) / 340.0) + 36.53
        self._debug_message("Returned temperature is %.2f°C" % temp)
        return(temp)

    def GetGyroscopeValues(self):
        """
        Read the gyroscope registers and return an x,y,z tuple
        """

        self._debug_message("Reading gyroscope registers")
        data = self._read_data(self.REG_GYRO_XOUT_H, 6)

        xout = self._two_comp((data[0] << 8) + data[1], 16) / self.gyro_lsb_sensitivity
        yout = self._two_comp((data[2] << 8) + data[3], 16) / self.gyro_lsb_sensitivity
        zout = self._two_comp((data[4] << 8) + data[5], 16) / self.gyro_lsb_sensitivity

        self._debug_message("MPU6050 gyroscope response :")
        self._debug_message("> X = %s°/s" % xout)
        self._debug_message("> Y = %s°/s" % yout)
        self._debug_message("> Z = %s°/s" % zout)

        return(xout, yout, zout)

    def GetAccelerometerValues(self, rotation=False):
        """
        Reads the accelerometer registers and return an x,y,z tuple

        If 'rotation' is set to True, this method will also determine
        the rotation angle on axes x and y and return this values on
        the tuple => x,y,z,rotx,roty (in degres)

        => Calculation method extracted from this article :
           http://www.hobbytronics.co.uk/accelerometer-info
        """

        self._debug_message("Reading accelerometer registers")
        data = self._read_data(self.REG_ACCEL_XOUT_H, 6)

        xout = self._two_comp((data[0] << 8) + data[1], 16) / self.accel_lsb_sensitivity
        yout = self._two_comp((data[2] << 8) + data[3], 16) / self.accel_lsb_sensitivity
        zout = self._two_comp((data[4] << 8) + data[5], 16) / self.accel_lsb_sensitivity

        self._debug_message("MPU6050 accelerometer response :")
        self._debug_message("> Accel X = %s" % xout)
        self._debug_message("> Accel Y = %s" % yout)
        self._debug_message("> Accel Z = %s" % zout)

        if rotation is True:
            try:
                xrot = math.degrees(math.atan(xout / math.sqrt((yout * yout) + (zout * zout))))
                yrot = math.degrees(math.atan(yout / math.sqrt((xout * xout) + (zout * zout))))

            except ZeroDivisionError:
                # Do not try to calculate rotation when registers are not inialized
                xrot = 0.0
                yrot = 0.0

            self._debug_message("> Rotation X = %s°" % xrot)
            self._debug_message("> Rotation Y = %s°" % yrot)
            return(xout, yout, zout, xrot, yrot)

        else:
            return(xout, yout, zout)

    def WhoAmI(self):
        """
        Reads the WHO_AM_I register that should return the I2C address of the chip
        """

        self._debug_message("Reading WHO_AM_I register")
        data = self._read_data(self.REG_WHO_AM_I, 1)
        return(data)
